/*

LSP-Fix (c) 2002-2006 Cexx.org

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


*/
//---------------------------------------------------------------------------
#include <vcl.h>
#include <Registry.hpp>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#pragma hdrstop

#include "Unit1.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TForm1 *Form1;

// Globals
AnsiString WS2Key="System\\CurrentControlSet\\Services\\Winsock2\\Parameters";
AnsiString key; // temporary variable for the current key

AnsiString Protocol;
AnsiString Namespace;



// Enum. Lists
TStringList* ProtocolList = new TStringList;
TStringList* NamespaceList = new TStringList;

//TStringList is just a list (array, in some non-C languages) of strings.


//---------------------------------------------------------------------------
__fastcall TForm1::TForm1(TComponent* Owner)
        : TForm(Owner)
{
        int ws2_errors=0;       // Error count for status-right-on-the-main-window

        // Setup
        TRegistry* Registry = new TRegistry;

	  // TRegistry is a Borland class that provides comparatively easy access to the Registry.
	  // TRegistry can only have one key 'open' at a time. Calls to it will return data for the currently opened key.
	  // Functions should be self-explanatory :)

        Registry->RootKey = HKEY_LOCAL_MACHINE;
	  // Sets the "Root key" for TRegsitry. This can be HKEY_LOCAL_MACHINE, HKEY_CURRENT_USER, etc.
	  // TRegistry will then treat this as the root for the entire registry, until you set the root to something else.

        // Check if Winsock2 keys exist
        if (!(Registry->KeyExists(WS2Key)))
        {
               AnsiString ErrorMsg="Winsock 2 Registry key (HKEY_LOCAL_MACHINE\\" + WS2Key + ") is missing or could not be accessed.\n\nIf using Windows NT/2000/XP, please make sure you are logged in as Administrator and try again.\n\nIf you are still receiving this message as Administrator, it may be necessary to re-install Winsock 2.";
               Application->MessageBox(ErrorMsg.c_str(),NULL, MB_OK);
               Form1->Close();
               exit(EXIT_FAILURE); // Exit the application returning a non-zero (failure) code
        }




        Registry->OpenKeyReadOnly(WS2Key);

	  // Reads the value for this name from the currently opened key.
        Protocol=Registry->ReadString("Current_Protocol_Catalog");
        Namespace=Registry->ReadString("Current_NameSpace_Catalog");
        Registry->CloseKey();


        if (Protocol=="" || Namespace=="")
        {
                // Bomb out if Winsock 2 keys missing
               Application->MessageBox("Protocol or Namespace Catalog keys not present in the Registry.\n\n Please re-install Winsock 2.",NULL, MB_OK);
               Form1->Close();
               exit(EXIT_FAILURE);

        }

        // Check Namespace Catalog
        key=WS2Key+"\\"+Namespace+"\\Catalog_Entries";
        if (!(Registry->KeyExists(key)))
        {
                // Debug
               AnsiString ErrorMsg="" + key + " does not exist.";
               Application->MessageBox(ErrorMsg.c_str(),NULL, MB_OK);
        }

        Registry->OpenKeyReadOnly(key);
        Registry->GetKeyNames(NamespaceList); // Return a list of data 'names' (from name-value pairs) into a TStringList
        Registry->CloseKey();
        NamespaceList->Sorted=true;     // in numerical key order

        // Check Protocol List
        key=WS2Key+"\\"+Protocol+"\\Catalog_Entries";
        if (!(Registry->KeyExists(key)))
        {
                // Debug
               AnsiString ErrorMsg="" + key + " does not exist.";
               Application->MessageBox(ErrorMsg.c_str(),NULL, MB_OK);
        }

        Registry->OpenKeyReadOnly(key);
        Registry->GetKeyNames(ProtocolList);
        Registry->CloseKey();
        ProtocolList->Sorted=true;      // in numerical key order


        // Now the fun begins.


        // First check that the total module counts are correct...
        Registry->OpenKeyReadOnly(WS2Key+"\\"+Namespace);
        if (Registry->ReadInteger("Num_Catalog_Entries") != NamespaceList->Count)
        {
                ws2_errors++;
        }
        Registry->CloseKey();

        Registry->OpenKeyReadOnly(WS2Key+"\\"+Protocol);
        if (Registry->ReadInteger("Num_Catalog_Entries") != ProtocolList->Count)
        {
                ws2_errors++;
        }
        Registry->CloseKey();

        // (easy) Get modules + descriptions from NameSpace list
	  // Note: NamespaceList->Count is the number of NamespaceList entries.
        for (int i=0; i<NamespaceList->Count; i++)
        {
                // Check that each NamespaceList subkey is in consecutive order
                if (NamespaceList->Strings[i].ToInt() != i+1)
                {
                        ws2_errors++;   // if any subkey(s) nonconsecutive
                }

                // Check that the referenced module actually exists
                key=WS2Key+"\\"+Namespace+"\\Catalog_Entries\\"+NamespaceList->Strings[i];
                Registry->OpenKeyReadOnly(key);
                AnsiString path=Registry->ReadString("LibraryPath");
                AnsiString desc=Registry->ReadString("DisplayString");
                Registry->CloseKey();
                path=GetRealPath(path);  // GetRealPath() resolves paths with variables (e.g. %windir%) into regular paths
                if (FileExists(path))
                {
                        AddToKeepList(path,desc);
                }
                else
                {
                        AddToDumpList(path,desc);
                        ws2_errors++;   // we found a nonexistent module
                }
        }

        // (harder) Get module names from Protocol list
        for (int i=0; i<ProtocolList->Count; i++)
        {
                // Check that each ProtocolList subkey is in consecutive order
                if (ProtocolList->Strings[i].ToInt() != i+1)
                {
                        ws2_errors++;   // if any subkey(s) nonconsecutive
                }

                // Check that the referenced module actually exists
                key=WS2Key+"\\"+Protocol+"\\Catalog_Entries\\"+ProtocolList->Strings[i];
                Registry->OpenKeyReadOnly(key);
                int bufSize=1024;
                char *Buffer=new char[bufSize]; // A potential problem if crapware vendors can safely insert a key with length >1024.

                Registry->ReadBinaryData("PackedCatalogItem", Buffer, bufSize);
                AnsiString desc="(Protocol handler)";

                AnsiString path=Buffer; // Terminates at the first 0x00
                Registry->CloseKey();
                path=GetRealPath(path);
                if (FileExists(path))
                {
                        AddToKeepList(path,desc);
                }
                else
                {
                        AddToDumpList(path,desc);
                        ws2_errors++;   // we found a nonexistent module
                }
                delete Buffer;

                // Ready to display WndMain... notify if any problems were found

                if (ws2_errors==0)
                {
                        StatusLabel->Caption="No problems found.";
                }
                else
                {
                        StatusLabel->Caption="Problems found in LSP chain. Press 'Finish' to make corrections. To exit without making changes, simply close the program.";
                        StatusLabel->Font->Color=clRed;
                }
        }
}
//---------------------------------------------------------------------------





//---------------------------------------------------------------------------

void __fastcall TForm1::FormCloseQuery(TObject *Sender, bool &CanClose)
{
        CanClose = true;
}
//---------------------------------------------------------------------------

// The following lists a file on the "Keep" list.
void TForm1::AddToKeepList(AnsiString path, AnsiString desc)
{
        int bsIndex=path.LastDelimiter("\\"); //index of the final backslash
        AnsiString file=path.SubString(bsIndex+1,(path.Length()+1 - bsIndex));
        int dupe=0;
        for (int i=0; i<KeepList->Items->Count; i++)
        {
		    // Make sure each item is only added to the lists once.
                if (file==KeepList->Items->Item[i]->Caption)
                {
                        dupe=1;
                }
        }
        if (dupe==0)
        {
                TListItem *pItem = KeepList->Items->Add (); // Adds one entry to the List (GUI element)
                pItem->Caption=file; // Adds the name of the item just added
                pItem->SubItems->Add(desc); // Adds the description for the item just added
        }
}

// The following lists a file on the "Remove" list.
void TForm1::AddToDumpList(AnsiString path, AnsiString desc)
{
        int bsIndex=path.LastDelimiter("\\"); //index of the final backslash
        AnsiString file=path.SubString(bsIndex+1,(path.Length()+1 - bsIndex));
        int dupe=0;
        for (int i=0; i<DumpList->Items->Count; i++)
        {
                if (file==DumpList->Items->Item[i]->Caption)
                {
                        dupe=1;
                }
        }
        if (dupe==0)
        {
                TListItem *pItem = DumpList->Items->Add ();
                pItem->Caption=file;
                pItem->SubItems->Add(desc);
        }
}


// The following takes a path containing environment variables (such as "%windir%") and resolves them into regular paths
// that can be used by anything.

AnsiString TForm1::GetRealPath(AnsiString path)
{
        int posn;
        int i=0;
        while (_environ[i]) // _environ[] is an array containing environment variables in the form "name=value"
        {
                AnsiString penvl=_environ[i];

		    // The following few lines split the environment line ("name=value") into envname and envvalue,
		    // and replace the variable with its value.

                posn=penvl.Pos("=");
                AnsiString envname=penvl.SubString(0,posn-1);
                envname="%" + envname.LowerCase() + "%";
                AnsiString envvalue=penvl.SubString(posn+1,penvl.Length()-posn);
                if (path.LowerCase().Pos(envname) > 0)
                {
                        path=Replace(path,envname,envvalue);
                }

                i++;
        }

        return path;
}

// The following is a portable "Replace" function used by GetRealPath. It replaces a substring
// with another substring.
AnsiString TForm1::Replace(AnsiString path, AnsiString from, AnsiString to)
{
        int start=path.LowerCase().Pos(from.LowerCase());
        int length=from.Length();
        path=path.SubString(0,start-1) + to + path.SubString(start+length,path.Length()-start-length+1);
        return path;
}

// The following transfers an item from the "Keep" to "Remove" list.
void __fastcall TForm1::ButtonRemoveClick(TObject *Sender)
{
        if (KeepList->Selected && CheckBox1->Checked)
        {
                TListItem *pSwap;
                TListItem *pItem=KeepList->Selected;
                pSwap=DumpList->Items->Add();
                pSwap->Caption=pItem->Caption;
                pSwap->SubItems->Add(pItem->SubItems->Strings[0]);
                pItem->Delete();
        }
}
//---------------------------------------------------------------------------

// The following transfers an item from the "Remove" to "Keep" list.
void __fastcall TForm1::ButtonKeepClick(TObject *Sender)
{
        if (DumpList->Selected && CheckBox1->Checked)
        {
                TListItem *pSwap;
                TListItem *pItem=DumpList->Selected;
                pSwap=KeepList->Items->Add();
                pSwap->Caption=pItem->Caption;
                pSwap->SubItems->Add(pItem->SubItems->Strings[0]);
                pItem->Delete();
        }
}
//---------------------------------------------------------------------------


// Used to enable/disable visual components (Keep/Remove buttons).
void __fastcall TForm1::CheckBox1Click(TObject *Sender)
{
        if (CheckBox1->Checked)
        {
                ButtonRemove->Enabled=true;
                ButtonKeep->Enabled=true;
        }
        else
        {
                ButtonRemove->Enabled=false;
                ButtonKeep->Enabled=false;
        }
}
//---------------------------------------------------------------------------

// The following does all the dirty work of removing the unwanted LSPs appearing on the "Remove" list
// and making sure all entries are consecutive.
void __fastcall TForm1::FinishButtonClick(TObject *Sender)
{
        AnsiString NamespaceRemoved=0; // Counters to keep track of how many of each were removed/renumbered
        AnsiString NamespaceRenumbered=0;
        AnsiString ProtocolRemoved=0;
        AnsiString ProtocolRenumbered=0;

        TRegistry* Registry=new TRegistry;
        Registry->RootKey = HKEY_LOCAL_MACHINE;
        char Buffer[1024];

        //Remove items appearing on the "Remove" list.
        int numEntries;

                // Remove 'NameSpace' items
        numEntries=NamespaceList->Count;
        for (int i=0; i<numEntries; i++)
        {
                for(int ii=0; ii<DumpList->Items->Count; ii++)
                {
                        key=WS2Key+"\\"+Namespace+"\\Catalog_Entries\\"+NamespaceList->Strings[i];
                        Registry->OpenKey(key,false);  // The 'false' means it is not OK to create a new key if it does not exist
                        AnsiString path=Registry->ReadString("LibraryPath");
                        Registry->CloseKey();

                        if (path.LowerCase().Pos(DumpList->Items->Item[ii]->Caption.LowerCase())!=0)
                        {
                                Registry->DeleteKey(key); // <--- The dangerous part :-)
										// Comment the above line out until sure there are no bugs...
                                NamespaceRemoved=NamespaceRemoved.ToInt() + 1;
                        }

                }
        }

                // Renumber 'NameSpace' entries
        key=WS2Key+"\\"+Namespace+"\\Catalog_Entries";
        Registry->OpenKeyReadOnly(key);
        NamespaceList->Sorted=true;  // Sort the list numerically
        Registry->GetKeyNames(NamespaceList);
        Registry->CloseKey();

        for (int i=0; i<NamespaceList->Count; i++)
        {
                key=WS2Key+"\\"+Namespace+"\\Catalog_Entries\\";//+NamespaceList->Strings[i];

			// Check that numbering of each entry is consecutive with no gaps
                if (NamespaceList->Strings[i].ToInt() != i+1)
                {
                        sprintf(Buffer, "%012u", i+1);  // Generate the next number, if the next entry is not consecutive
                        Registry->MoveKey(key+NamespaceList->Strings[i], key+Buffer, true); // Move (rename) the key
                        NamespaceRenumbered=NamespaceRenumbered.ToInt()+1;
                }
        }

		// All finished removing and renumbering NameSpace items. Now update the total count in the Registry...
        Registry->OpenKey(WS2Key+"\\"+Namespace, false);
        Registry->WriteInteger("Num_Catalog_Entries",NamespaceList->Count);
        Registry->CloseKey();

		// All finished fixing NameSpace items.



                // Remove 'Protocol' items
        numEntries=ProtocolList->Count;
        for (int i=0; i<numEntries; i++)
        {
                for(int ii=0; ii<(DumpList->Items->Count); ii++)
                {
                        key=WS2Key+"\\"+Protocol+"\\Catalog_Entries\\"+ProtocolList->Strings[i];
                        Registry->OpenKey(key,false);
                        Registry->ReadBinaryData("PackedCatalogItem", Buffer, 1024);
                        AnsiString path=Buffer; // Terminates at the first '0'
                        Registry->CloseKey();

                        if (path.Length()>0)
                        {

                                if (path.LowerCase().Pos(DumpList->Items->Item[ii]->Caption.LowerCase())!=0)
                                {

                                        Registry->DeleteKey(key);  // Again, line where all the danger is.
											 // You have tested this before uncommenting this line, right? :-)
                                        ProtocolRemoved=ProtocolRemoved.ToInt()+1;
                                }
                        }
                }
        }

	  // Finished removing 'Protocol' items


                // Renumber 'Protocol' entries
        key=WS2Key+"\\"+Protocol+"\\Catalog_Entries";
        Registry->OpenKeyReadOnly(key);
        ProtocolList->Sorted=true;
        Registry->GetKeyNames(ProtocolList);
        Registry->CloseKey();

        for (int i=0; i<ProtocolList->Count; i++)
        {
                key=WS2Key+"\\"+Protocol+"\\Catalog_Entries\\";
                if (ProtocolList->Strings[i].ToInt() != i+1)
                {
                        sprintf(Buffer, "%012u", i+1);
                        Registry->MoveKey(key+ProtocolList->Strings[i], key+Buffer, true);
                        ProtocolRenumbered=ProtocolRenumbered.ToInt()+1;
                }
        }

	  // Finished removing and renumbering Protocol items.

        Registry->OpenKey(WS2Key+"\\"+Protocol, false);
        Registry->WriteInteger("Num_Catalog_Entries",ProtocolList->Count); // Update the count in the Registry
        Registry->CloseKey();
        AnsiString msg;

        // I don't for the life of me know why these counters were implemented as strings.
        // Possibly a braindead workaround for a braindead error message encountered
        // during a latenight caffeine-fueled hack session with the voices in my head!

        if (NamespaceRemoved.ToInt() + NamespaceRenumbered.ToInt() + ProtocolRemoved.ToInt() + ProtocolRenumbered.ToInt() == 0)
        {
                msg = "No changes necessary.\n\n";
        }
        else
        {
                msg = "Repairs complete.\n\n";
        }

        msg = msg + NamespaceRemoved + " NameSpace provider entries removed\n" + NamespaceRenumbered + " NameSpace provider entries renumbered\n" + ProtocolRemoved + " Protocol provider entries removed\n" + ProtocolRenumbered + " Protocol provider entries renumbered";

        Application->MessageBox(msg.c_str(),"Repair summary", MB_OK);
        Form1->Close(); // All done!
}

//---------------------------------------------------------------------------






